// From http://www.cr0.net:8040/code/crypto/aes/ but modified to fit into a C++ class and reduce to the minimum required.

/*
 *  FIPS-197 compliant AES implementation
 *
 *  Copyright (C) 2001-2004  Christophe Devine
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "stdafx.h"
#include "aes.h"


CCryptAES::CCryptAES()
{
	aes_gen_tables();
	aes_set_key_eoa();
}

/* forward S-box & tables */

uint32 FSb[256];
uint32 FT0[256]; 
uint32 FT1[256]; 
uint32 FT2[256]; 
uint32 FT3[256]; 

/* reverse S-box & tables */

uint32 RSb[256];
uint32 RT0[256];
uint32 RT1[256];
uint32 RT2[256];
uint32 RT3[256];

/* round constants */

uint32 RCON[10];

/* tables generation routine */

#define ROTR8(x) ( ( ( x << 24 ) & 0xFFFFFFFF ) | \
                   ( ( x & 0xFFFFFFFF ) >>  8 ) )

#define XTIME(x) ( ( x <<  1 ) ^ ( ( x & 0x80 ) ? 0x1B : 0x00 ) )
#define MUL(x,y) ( ( x &&  y ) ? pow[(log[x] + log[y]) % 255] : 0 )


void CCryptAES::aes_gen_tables( void )
{
    int i;
    uint8 x, y;
    uint8 pow[256];
    uint8 log[256];

    /* compute pow and log tables over GF(2^8) */

    for( i = 0, x = 1; i < 256; i++, x ^= XTIME( x ) )
    {
        pow[i] = x;
        log[x] = i;
    }

    /* calculate the round constants */

    for( i = 0, x = 1; i < 10; i++, x = XTIME( x ) )
    {
        RCON[i] = (uint32) x << 24;
    }

    /* generate the forward and reverse S-boxes */

    FSb[0x00] = 0x63;
    RSb[0x63] = 0x00;

    for( i = 1; i < 256; i++ )
    {
        x = pow[255 - log[i]];

        y = x;  y = ( y << 1 ) | ( y >> 7 );
        x ^= y; y = ( y << 1 ) | ( y >> 7 );
        x ^= y; y = ( y << 1 ) | ( y >> 7 );
        x ^= y; y = ( y << 1 ) | ( y >> 7 );
        x ^= y ^ 0x63;

        FSb[i] = x;
        RSb[x] = i;
    }

    /* generate the forward and reverse tables */

    for( i = 0; i < 256; i++ )
    {
        x = (unsigned char) FSb[i]; y = XTIME( x );

        FT0[i] =   (uint32) ( x ^ y ) ^
                 ( (uint32) x <<  8 ) ^
                 ( (uint32) x << 16 ) ^
                 ( (uint32) y << 24 );

        FT0[i] &= 0xFFFFFFFF;

        FT1[i] = ROTR8( FT0[i] );
        FT2[i] = ROTR8( FT1[i] );
        FT3[i] = ROTR8( FT2[i] );

        y = (unsigned char) RSb[i];

        RT0[i] = ( (uint32) MUL( 0x0B, y )       ) ^
                 ( (uint32) MUL( 0x0D, y ) <<  8 ) ^
                 ( (uint32) MUL( 0x09, y ) << 16 ) ^
                 ( (uint32) MUL( 0x0E, y ) << 24 );

        RT0[i] &= 0xFFFFFFFF;

        RT1[i] = ROTR8( RT0[i] );
        RT2[i] = ROTR8( RT1[i] );
        RT3[i] = ROTR8( RT2[i] );
    }
}


/* platform-independant 32-bit integer manipulation macros */

#define GET_UINT32(n,b,i)                       \
{                                               \
    (n) = ( (uint32) (b)[(i)    ] << 24 )       \
        | ( (uint32) (b)[(i) + 1] << 16 )       \
        | ( (uint32) (b)[(i) + 2] <<  8 )       \
        | ( (uint32) (b)[(i) + 3]       );      \
}

#define PUT_UINT32(n,b,i)                       \
{                                               \
    (b)[(i)    ] = (uint8) ( (n) >> 24 );       \
    (b)[(i) + 1] = (uint8) ( (n) >> 16 );       \
    (b)[(i) + 2] = (uint8) ( (n) >>  8 );       \
    (b)[(i) + 3] = (uint8) ( (n)       );       \
}


// This setup the right key for End of Age (I only have the 'expanded' key, not the 'real' 128 bits key) but no need
// for it once I have the expanded one.
void CCryptAES::aes_set_key_eoa()
{
	nr = 10;

	// This taken from EoA Demo program directly from memory (on my system, located at 0x8f63f0) and reverse so that it is
	// compatible with the way the decryption works in this module. (last row becomes first row and so on)
	uint8 expDecrypt[176] =	{ 0xA0, 0x55, 0x17, 0xBD, 0x20, 0x26, 0x3B, 0x8E, 0xDA, 0x59, 0x9B, 0x0E, 0xAE, 0x9D, 0xD2, 0x59,
														0xBF, 0xE6, 0x6C, 0xB2, 0x36, 0x1B, 0x59, 0x98, 0x62, 0xF1, 0x69, 0x5F, 0x52, 0x16, 0x28, 0xC2,
														0x23, 0xBD, 0xE4, 0x05, 0x89, 0xFD, 0x35, 0x2A, 0x54, 0xEA, 0x30, 0xC7, 0x30, 0xE7, 0x41, 0x9D,
														0x45, 0x3D, 0xC3, 0x5C, 0xAA, 0x40, 0xD1, 0x2F, 0xDD, 0x17, 0x05, 0xED, 0x64, 0x0D, 0x71, 0x5A,
														0xBC, 0xCB, 0x5C, 0x4F, 0xEF, 0x7D, 0x12, 0x73, 0x77, 0x57, 0xD4, 0xC2, 0xB9, 0x1A, 0x74, 0xB7,
														0x3F, 0x88, 0x59, 0xE5, 0x53, 0xB6, 0x4E, 0x3C, 0x98, 0x2A, 0xC6, 0xB1, 0xCE, 0x4D, 0xA0, 0x75,
														0x48, 0x9E, 0xEC, 0xD6, 0x6C, 0x3E, 0x17, 0xD9, 0xCB, 0x9C, 0x88, 0x8D, 0x56, 0x67, 0x66, 0xC4,
														0x1C, 0xCC, 0xB7, 0x11, 0x24, 0xA0, 0xFB, 0x0F, 0xA7, 0xA2, 0x9F, 0x54, 0x9D, 0xFB, 0xEE, 0x49,
														0x25, 0xD8, 0xBF, 0x3C, 0x38, 0x6C, 0x4C, 0x1E, 0x83, 0x02, 0x64, 0x5B, 0x3A, 0x59, 0x71, 0x1D,
														0xD2, 0x81, 0xB4, 0x13, 0x1D, 0xB4, 0xF3, 0x22, 0xBB, 0x6E, 0x28, 0x45, 0xB9, 0x5B, 0x15, 0x46,
														0xF0, 0x4D, 0x25, 0x33, 0xAC, 0x5D, 0x27, 0x5A, 0x9E, 0x18, 0x78, 0x3E, 0x65, 0x2C, 0x48, 0x08 };


	// Need to reverse the 'endian' so that it is compatible with the way this module works (this module is little
	// endian while the algorithm used in EoA seem to be big endian or the opposite...)
	memset(drk, 0, 64 * 4);
	for (int index = 0; index < 176; index += 4)
    GET_UINT32(drk[index >> 2], expDecrypt,  index);



	// Same stuff but for encryption (no need to reverse the row in this case)
	uint8 expEncrypt[176] =	{ 0xF0, 0x4D, 0x25, 0x33, 0xAC, 0x5D, 0x27, 0x5A, 0x9E, 0x18, 0x78, 0x3E, 0x65, 0x2C, 0x48, 0x08,
														0x80, 0x1F, 0x15, 0x7E, 0x2C, 0x42, 0x32, 0x24, 0xB2, 0x5A, 0x4A, 0x1A, 0xD7, 0x76, 0x02, 0x12,
														0xBA, 0x68, 0xDC, 0x70, 0x96, 0x2A, 0xEE, 0x54, 0x24, 0x70, 0xA4, 0x4E, 0xF3, 0x06, 0xA6, 0x5C,
														0xD1, 0x4C, 0x96, 0x7D, 0x47, 0x66, 0x78, 0x29, 0x63, 0x16, 0xDC, 0x67, 0x90, 0x10, 0x7A, 0x3B,
														0x13, 0x96, 0x74, 0x1D, 0x54, 0xF0, 0x0C, 0x34, 0x37, 0xE6, 0xD0, 0x53, 0xA7, 0xF6, 0xAA, 0x68,
														0x41, 0x3A, 0x31, 0x41, 0x15, 0xCA, 0x3D, 0x75, 0x22, 0x2C, 0xED, 0x26, 0x85, 0xDA, 0x47, 0x4E,
														0x36, 0x9A, 0x1E, 0xD6, 0x23, 0x50, 0x23, 0xA3, 0x01, 0x7C, 0xCE, 0x85, 0x84, 0xA6, 0x89, 0xCB,
														0x52, 0x3D, 0x01, 0x89, 0x71, 0x6D, 0x22, 0x2A, 0x70, 0x11, 0xEC, 0xAF, 0xF4, 0xB7, 0x65, 0x64,
														0x7B, 0x70, 0x42, 0x36, 0x0A, 0x1D, 0x60, 0x1C, 0x7A, 0x0C, 0x8C, 0xB3, 0x8E, 0xBB, 0xE9, 0xD7,
														0x8A, 0x6E, 0x4C, 0x2F, 0x80, 0x73, 0x2C, 0x33, 0xFA, 0x7F, 0xA0, 0x80, 0x74, 0xC4, 0x49, 0x57,
														0xA0, 0x55, 0x17, 0xBD, 0x20, 0x26, 0x3B, 0x8E, 0xDA, 0x59, 0x9B, 0x0E, 0xAE, 0x9D, 0xD2, 0x59 };


	memset(erk, 0, 64 * 4);
	for (index = 0; index < 176; index += 4)
    GET_UINT32(erk[index >> 2], expEncrypt,  index);
}


/* AES 128-bit block encryption routine */

void CCryptAES::aes_encrypt( uint8 input[16], uint8 output[16] )
{
    uint32 *RK, X0, X1, X2, X3, Y0, Y1, Y2, Y3;

    RK = erk;

    GET_UINT32( X0, input,  0 );
    GET_UINT32( X1, input,  4 );
    GET_UINT32( X2, input,  8 );
    GET_UINT32( X3, input, 12 );

		X0 ^= RK[0];
		X1 ^= RK[1];
		X2 ^= RK[2];
		X3 ^= RK[3];

#define AES_FROUND(X0,X1,X2,X3,Y0,Y1,Y2,Y3)     \
{                                               \
    RK += 4;                                    \
                                                \
    X0 = RK[0] ^ FT0[ (uint8) ( Y0 >> 24 ) ] ^  \
                 FT1[ (uint8) ( Y1 >> 16 ) ] ^  \
                 FT2[ (uint8) ( Y2 >>  8 ) ] ^  \
                 FT3[ (uint8) ( Y3       ) ];   \
                                                \
    X1 = RK[1] ^ FT0[ (uint8) ( Y1 >> 24 ) ] ^  \
                 FT1[ (uint8) ( Y2 >> 16 ) ] ^  \
                 FT2[ (uint8) ( Y3 >>  8 ) ] ^  \
                 FT3[ (uint8) ( Y0       ) ];   \
                                                \
    X2 = RK[2] ^ FT0[ (uint8) ( Y2 >> 24 ) ] ^  \
                 FT1[ (uint8) ( Y3 >> 16 ) ] ^  \
                 FT2[ (uint8) ( Y0 >>  8 ) ] ^  \
                 FT3[ (uint8) ( Y1       ) ];   \
                                                \
    X3 = RK[3] ^ FT0[ (uint8) ( Y3 >> 24 ) ] ^  \
                 FT1[ (uint8) ( Y0 >> 16 ) ] ^  \
                 FT2[ (uint8) ( Y1 >>  8 ) ] ^  \
                 FT3[ (uint8) ( Y2       ) ];   \
}

    AES_FROUND( Y0, Y1, Y2, Y3, X0, X1, X2, X3 );       /* round 1 */
    AES_FROUND( X0, X1, X2, X3, Y0, Y1, Y2, Y3 );       /* round 2 */
    AES_FROUND( Y0, Y1, Y2, Y3, X0, X1, X2, X3 );       /* round 3 */
    AES_FROUND( X0, X1, X2, X3, Y0, Y1, Y2, Y3 );       /* round 4 */
    AES_FROUND( Y0, Y1, Y2, Y3, X0, X1, X2, X3 );       /* round 5 */
    AES_FROUND( X0, X1, X2, X3, Y0, Y1, Y2, Y3 );       /* round 6 */
    AES_FROUND( Y0, Y1, Y2, Y3, X0, X1, X2, X3 );       /* round 7 */
    AES_FROUND( X0, X1, X2, X3, Y0, Y1, Y2, Y3 );       /* round 8 */
    AES_FROUND( Y0, Y1, Y2, Y3, X0, X1, X2, X3 );       /* round 9 */

    if( nr > 10 )
    {
        AES_FROUND( X0, X1, X2, X3, Y0, Y1, Y2, Y3 );   /* round 10 */
        AES_FROUND( Y0, Y1, Y2, Y3, X0, X1, X2, X3 );   /* round 11 */
    }

    if( nr > 12 )
    {
        AES_FROUND( X0, X1, X2, X3, Y0, Y1, Y2, Y3 );   /* round 12 */
        AES_FROUND( Y0, Y1, Y2, Y3, X0, X1, X2, X3 );   /* round 13 */
    }

    /* last round */

    RK += 4;

    X0 = RK[0] ^ ( FSb[ (uint8) ( Y0 >> 24 ) ] << 24 ) ^
                 ( FSb[ (uint8) ( Y1 >> 16 ) ] << 16 ) ^
                 ( FSb[ (uint8) ( Y2 >>  8 ) ] <<  8 ) ^
                 ( FSb[ (uint8) ( Y3       ) ]       );

    X1 = RK[1] ^ ( FSb[ (uint8) ( Y1 >> 24 ) ] << 24 ) ^
                 ( FSb[ (uint8) ( Y2 >> 16 ) ] << 16 ) ^
                 ( FSb[ (uint8) ( Y3 >>  8 ) ] <<  8 ) ^
                 ( FSb[ (uint8) ( Y0       ) ]       );

    X2 = RK[2] ^ ( FSb[ (uint8) ( Y2 >> 24 ) ] << 24 ) ^
                 ( FSb[ (uint8) ( Y3 >> 16 ) ] << 16 ) ^
                 ( FSb[ (uint8) ( Y0 >>  8 ) ] <<  8 ) ^
                 ( FSb[ (uint8) ( Y1       ) ]       );

    X3 = RK[3] ^ ( FSb[ (uint8) ( Y3 >> 24 ) ] << 24 ) ^
                 ( FSb[ (uint8) ( Y0 >> 16 ) ] << 16 ) ^
                 ( FSb[ (uint8) ( Y1 >>  8 ) ] <<  8 ) ^
                 ( FSb[ (uint8) ( Y2       ) ]       );

    PUT_UINT32( X0, output,  0 );
    PUT_UINT32( X1, output,  4 );
    PUT_UINT32( X2, output,  8 );
    PUT_UINT32( X3, output, 12 );
}

/* AES 128-bit block decryption routine */

void CCryptAES::aes_decrypt( uint8 input[16], uint8 output[16] )
{
    uint32 *RK, X0, X1, X2, X3, Y0, Y1, Y2, Y3;

    RK = drk;

    GET_UINT32( X0, input,  0 );
    GET_UINT32( X1, input,  4 );
    GET_UINT32( X2, input,  8 );
    GET_UINT32( X3, input, 12 );

		X0 ^= RK[0];
		X1 ^= RK[1];
		X2 ^= RK[2];
		X3 ^= RK[3];

#define AES_RROUND(X0,X1,X2,X3,Y0,Y1,Y2,Y3)     \
{                                               \
    RK += 4;                                    \
                                                \
    X0 = RK[0] ^ RT0[ (uint8) ( Y0 >> 24 ) ] ^  \
                 RT1[ (uint8) ( Y3 >> 16 ) ] ^  \
                 RT2[ (uint8) ( Y2 >>  8 ) ] ^  \
                 RT3[ (uint8) ( Y1       ) ];   \
                                                \
    X1 = RK[1] ^ RT0[ (uint8) ( Y1 >> 24 ) ] ^  \
                 RT1[ (uint8) ( Y0 >> 16 ) ] ^  \
                 RT2[ (uint8) ( Y3 >>  8 ) ] ^  \
                 RT3[ (uint8) ( Y2       ) ];   \
                                                \
    X2 = RK[2] ^ RT0[ (uint8) ( Y2 >> 24 ) ] ^  \
                 RT1[ (uint8) ( Y1 >> 16 ) ] ^  \
                 RT2[ (uint8) ( Y0 >>  8 ) ] ^  \
                 RT3[ (uint8) ( Y3       ) ];   \
                                                \
    X3 = RK[3] ^ RT0[ (uint8) ( Y3 >> 24 ) ] ^  \
                 RT1[ (uint8) ( Y2 >> 16 ) ] ^  \
                 RT2[ (uint8) ( Y1 >>  8 ) ] ^  \
                 RT3[ (uint8) ( Y0       ) ];   \
}

    AES_RROUND( Y0, Y1, Y2, Y3, X0, X1, X2, X3 );       /* round 1 */
    AES_RROUND( X0, X1, X2, X3, Y0, Y1, Y2, Y3 );       /* round 2 */
    AES_RROUND( Y0, Y1, Y2, Y3, X0, X1, X2, X3 );       /* round 3 */
    AES_RROUND( X0, X1, X2, X3, Y0, Y1, Y2, Y3 );       /* round 4 */
    AES_RROUND( Y0, Y1, Y2, Y3, X0, X1, X2, X3 );       /* round 5 */
    AES_RROUND( X0, X1, X2, X3, Y0, Y1, Y2, Y3 );       /* round 6 */
    AES_RROUND( Y0, Y1, Y2, Y3, X0, X1, X2, X3 );       /* round 7 */
    AES_RROUND( X0, X1, X2, X3, Y0, Y1, Y2, Y3 );       /* round 8 */
    AES_RROUND( Y0, Y1, Y2, Y3, X0, X1, X2, X3 );       /* round 9 */

    if( nr > 10 )
    {
        AES_RROUND( X0, X1, X2, X3, Y0, Y1, Y2, Y3 );   /* round 10 */
        AES_RROUND( Y0, Y1, Y2, Y3, X0, X1, X2, X3 );   /* round 11 */
    }

    if( nr > 12 )
    {
        AES_RROUND( X0, X1, X2, X3, Y0, Y1, Y2, Y3 );   /* round 12 */
        AES_RROUND( Y0, Y1, Y2, Y3, X0, X1, X2, X3 );   /* round 13 */
    }

    /* last round */

    RK += 4;

    X0 = RK[0] ^ ( RSb[ (uint8) ( Y0 >> 24 ) ] << 24 ) ^
                 ( RSb[ (uint8) ( Y3 >> 16 ) ] << 16 ) ^
                 ( RSb[ (uint8) ( Y2 >>  8 ) ] <<  8 ) ^
                 ( RSb[ (uint8) ( Y1       ) ]       );

    X1 = RK[1] ^ ( RSb[ (uint8) ( Y1 >> 24 ) ] << 24 ) ^
                 ( RSb[ (uint8) ( Y0 >> 16 ) ] << 16 ) ^
                 ( RSb[ (uint8) ( Y3 >>  8 ) ] <<  8 ) ^
                 ( RSb[ (uint8) ( Y2       ) ]       );

    X2 = RK[2] ^ ( RSb[ (uint8) ( Y2 >> 24 ) ] << 24 ) ^
                 ( RSb[ (uint8) ( Y1 >> 16 ) ] << 16 ) ^
                 ( RSb[ (uint8) ( Y0 >>  8 ) ] <<  8 ) ^
                 ( RSb[ (uint8) ( Y3       ) ]       );

    X3 = RK[3] ^ ( RSb[ (uint8) ( Y3 >> 24 ) ] << 24 ) ^
                 ( RSb[ (uint8) ( Y2 >> 16 ) ] << 16 ) ^
                 ( RSb[ (uint8) ( Y1 >>  8 ) ] <<  8 ) ^
                 ( RSb[ (uint8) ( Y0       ) ]       );

    PUT_UINT32( X0, output,  0 );
    PUT_UINT32( X1, output,  4 );
    PUT_UINT32( X2, output,  8 );
    PUT_UINT32( X3, output, 12 );
}

// Decode the 'buffer' using the EoA default 128 bits key.  The 'lenBuffer' must be on an 16 bytes boundary.
// Returns FALSE in case of error.  Buffer decrypted 'in-place'.
BOOL CCryptAES::decodeBuffer(uint8 *buffer, int lenBuffer)
{
	// Make sure parameter are valid
	if (!buffer || (lenBuffer <= 0) || ((lenBuffer % 16) != 0))
		return FALSE;

	int nbrSeq = lenBuffer >> 4;
	for (int index = 0; index < nbrSeq; index++)
		aes_decrypt(buffer + (16 * index), buffer + (16 * index));
	return TRUE;
}


// Encode the 'buffer' using the EoA default 128 bits key.  The 'lenBuffer' must be on an 16 bytes boundary.
// Returns FALSE in case of error.  Buffer decrypted 'in-place'.
BOOL CCryptAES::encodeBuffer(uint8 *buffer, int lenBuffer)
{
	// Make sure parameter are valid
	if (!buffer || (lenBuffer <= 0) || ((lenBuffer % 16) != 0))
		return FALSE;

	int nbrSeq = lenBuffer >> 4;
	for (int index = 0; index < nbrSeq; index++)
		aes_encrypt(buffer + (16 * index), buffer + (16 * index));
	return TRUE;
}

